--
-- gemooy.lua
-- Gemooy interpreter in Lua
--
-- SPDX-FileCopyrightText: In 2024, Chris Pressey, the original author of this work, placed it into the public domain.
-- For more information, please refer to <https://unlicense.org/>
-- SPDX-License-Identifier: Unlicense
--

local Debugging = require "debugging"
local Playfield = require "playfield"

--
-- Utils
--

local rotate_cw = function(dx, dy)
    if dx == 0 and dy == -1 then
        return 1, -1
    elseif dx == 1 and dy == -1 then
        return 1, 0
    elseif dx == 1 and dy == 0 then
        return 1, 1
    elseif dx == 1 and dy == 1 then
        return 0, 1
    elseif dx == 0 and dy == 1 then
        return -1, 1
    elseif dx == -1 and dy == 1 then
        return -1, 0
    elseif dx == -1 and dy == 0 then
        return -1, -1
    elseif dx == -1 and dy == -1 then
        return 0, -1
    end
end

local rotate_ccw = function(dx, dy)
    if dx == 0 and dy == -1 then
        return -1, -1
    elseif dx == -1 and dy == -1 then
        return -1, 0
    elseif dx == -1 and dy == 0 then
        return -1, 1
    elseif dx == -1 and dy == 1 then
        return 0, 1
    elseif dx == 0 and dy == 1 then
        return 1, 1
    elseif dx == 1 and dy == 1 then
        return 1, 0
    elseif dx == 1 and dy == 0 then
        return 1, -1
    elseif dx == 1 and dy == -1 then
        return 0, -1
    end
end

local increment_playfield_at = function(p, x, y)
    local data = p:peek(x, y)
    if data == " " then
        data = "#"
    elseif data == "#" then
        data = "@"
    elseif data == "@" then
        data = " "
    else
        error("What is this i cannot even " + data)
    end
    p:poke(x, y, data)
end

local decrement_playfield_at = function(p, x, y)
    local data = p:peek(x, y)
    if data == " " then
        data = "@"
    elseif data == "@" then
        data = "#"
    elseif data == "#" then
        data = " "
    else
        error("What is this i cannot even " + data)
    end
    p:poke(x, y, data)
end

--
-- Loader
--

local load_file = function(filename)
    local f = assert(io.open(filename, "r"))
    local program_text = f:read("*all")
    f:close()
    return program_text
end

--
-- Visualization
--

local draw_playfield = function(p, ix, iy, px, py)
    local wrapper = function(c, x, y)
        if x == ix and y == iy then
            return "\027[41m" .. c .. "\027[0m"
        elseif x == px and y == py then
            return "\027[42m" .. c .. "\027[0m"
        else
            return c
        end
    end
    io.write(p:render(wrapper))
end

--
-- Main
--

function main(arg)
    math.randomseed(os.time())
    local filename
    local headless = false
    local sleep_command = false
    while #arg > 0 do
        if arg[1] == "--debug" then
            table.remove(arg, 1)
            Debugging.debug_these(arg[1])
        elseif arg[1] == "--headless" then
            headless = true
        elseif arg[1] == "--sleep-command" then
            table.remove(arg, 1)
            sleep_command = arg[1]
        elseif arg[1] ~= "" then
            filename = arg[1]
        end
        table.remove(arg, 1)
    end
    local program_text = load_file(filename)
    local p = Playfield.new()
    local dx = 1
    local dy = 1
    local ix, iy, px, py
    p:load(program_text, function(c, x, y)
        if c == "%" then
            px = x
            py = y
        elseif c == "$" then
            ix = x
            iy = y
        elseif c == "#" or c == "@" then
            return c
        end
    end)

    local done = false

    if not headless then
        io.write("\027[H\027[2J")  -- clear screen
        draw_playfield(p)
    end
    if sleep_command then
        local result = os.execute(sleep_command)
        if not result then
            done = true
        end
    end

    while not done do
        local instr = p:peek(ix, iy)
        local data = p:peek(px, py)
        Debugging.debug(
            "state", "ix: " .. ix .. ", iy: " .. iy .. ", instr: '" .. instr .. "', " ..
                     "px: " .. px .. ", py: " .. py .. ", data: " .. "'" .. data .. "'"
        )
        if instr == "@" then
            Debugging.debug("state", "rotate begin, dx: " .. dx .. ", dy: " .. dy .. ", data: '" .. data .. "'")
            if data == " " then
                dx, dy = rotate_cw(dx, dy)
            elseif data == "#" then
                dx, dy = rotate_ccw(dx, dy)
            end
            Debugging.debug("state", "rotate end, dx: " .. dx .. ", dy: " .. dy)
        elseif instr == "#" then
            if dx == 0 and dy == -1 then
                py = py - 1
                ix = ix + dx
                iy = iy + dy
            elseif dx == 0 and dy == 1 then
                py = py + 1
                ix = ix + dx
                iy = iy + dy
            elseif dx == 1 and dy == 0 then
                px = px + 1
                ix = ix + dx
                iy = iy + dy
            elseif dx == -1 and dy == 0 then
                px = px - 1
                ix = ix + dx
                iy = iy + dy
            elseif (dx == -1 and dy == -1) or (dx == 1 and dy == -1) then
                increment_playfield_at(p, px, py)
            elseif (dx == -1 and dy == 1) or (dx == 1 and dy == 1) then
                decrement_playfield_at(p, px, py)
            end
        end

        ix = ix + dx
        iy = iy + dy

        if not headless then
            io.write("\027[H")
            draw_playfield(p, ix, iy, px, py)
        end
        if sleep_command then
            local result = os.execute(sleep_command)
            if not result then
                done = true
            end
        end
    end
end

if arg ~= nil then
    main(arg)
end
